using SparseArrays: sparse
using LinearAlgebra: det, dot
using StaticArrays: SA, SArray, MArray
using ForwardDiff: jacobian

export Poisson, L2Projection, init_point!, assemble, project_img, project_img2

abstract type Operator end

Base.show(io::IO, ::MIME"text/plain", x::Operator) =
    print("$(nameof(typeof(x)))")


struct L2Projection{Sp, F} <: Operator
    space::Sp
    f::F # target
end

L2Projection(f) = L2Projection(f.space, f)

params(op::L2Projection) = (f = op.f,)
a(op::L2Projection, xloc, u, du, v, dv; f) = dot(u, v)
l(op::L2Projection, xloc, v, dv; f) = dot(f, v)



struct Poisson{Sp} <: Operator
    space::Sp
end

a(op::Poisson, xloc, u, du, v, dv) = dot(du, dv)



# degree 2 quadrature on triangle
quadrature() = SA[1/6, 1/6, 1/6], SA[1/6 4/6 1/6; 1/6 1/6 4/6]
# degree 1 quadrature on triangle using corner points
quadrature_lumped() = SA[1/6, 1/6, 1/6], SA[0. 0. 1.; 0. 1. 0.]


assemble(op::Operator) = assemble(op.space,
    (x...; y...) -> a(op, x...; y...),
    (x...; y...) -> l(op, x...; y...);
    params(op)...)

assemble(space::FeSpace, a, l; params...) =
    assemble(space::FeSpace, quadrature(), a, l; params...)

assemble_lumped(space::FeSpace, a, l; params...) =
    assemble(space::FeSpace, quadrature_lumped(), a, l; params...)

function assemble(space::FeSpace, quadrature, a, l; params...)
    mesh = space.mesh
    opparams = NamedTuple(params)

    # precompute basis at quadrature points

    qw, qx = quadrature

    d = size(qx, 1) # domain dimension
    nrdims = prod(space.size)
    nldofs = ndofs(space.element) # number of element dofs (i.e. local dofs not counting codomain dimensions)
    nqpts = length(qw) # number of quadrature points

    qphi_ = zeros(nrdims, nrdims, nldofs, nqpts)
    dqphi_ = zeros(nrdims, d, nrdims, nldofs, nqpts)
    for r in 1:nrdims
	for k in axes(qx, 2)
	    qphi_[r, r, :, k] .= evaluate_basis(space.element, qx[:, k])
	    dqphi_[r, :, r, :, k] .= transpose(jacobian(x -> evaluate_basis(space.element, x), qx[:, k]))
	end
    end
    qphi = SArray{Tuple{nrdims, nrdims, nldofs, nqpts}}(qphi_)
    dqphi = SArray{Tuple{nrdims, d, nrdims, nldofs, nqpts}}(dqphi_)

    n = ncells(mesh) * size(qx, 2) *
        nrdims * nldofs * nrdims * nldofs
    I = Vector{Int}(undef, n)
    J = Vector{Int}(undef, n)
    V = Vector{Float64}(undef, n)
    spidx = 0
    b = zeros(ndofs(space))
    gdof = LinearIndices((nrdims, ndofs(space)))

    # mesh cells
    for cell in cells(mesh)
	foreach(f -> bind!(f, cell), opparams)
	# cell map is assumed to be constant per cell
	delmap = jacobian(elmap(mesh, cell), SA[0., 0.])
	delmapinv = inv(delmap)
	intel = abs(det(delmap))

	# quadrature points
	for k in axes(qx, 2)
	    xhat = qx[:, k]::SArray
	    x = elmap(mesh, cell)(xhat)
	    opvalues = map(f -> evaluate(f, xhat), opparams)

	    # local test-function dofs
	    for jdim in 1:nrdims, ldofj in 1:nldofs
		gdofj = space.dofmap[jdim, ldofj, cell]

		phij = SArray{Tuple{space.size...}}(qphi[:, jdim, ldofj, k])
		dphij = SArray{Tuple{space.size..., d}}(dqphi[:, :, jdim, ldofj, k] * delmapinv)

		lv = qw[k] * l(x, phij, dphij; opvalues...) * intel
		b[gdofj] += lv

		# local trial-function dofs
		for idim in 1:nrdims, ldofi in 1:nldofs
		    gdofi = space.dofmap[idim, ldofi, cell]

		    phii = SArray{Tuple{space.size...}}(qphi[:, idim, ldofi, k])
		    dphii = SArray{Tuple{space.size..., d}}(dqphi[:, :, idim, ldofi, k] * delmapinv)

		    av = qw[k] * a(x, phii, dphii, phij, dphij; opvalues...) * intel
                    spidx += 1
                    I[spidx] = gdofi
                    J[spidx] = gdofj
                    V[spidx] = av
		end
	    end
	end
    end
    @assert spidx == n

    ngdofs = ndofs(space)
    A = sparse(I, J, V, ngdofs, ngdofs)
    return A, b
end

# composite midpoint quadrature on 2d-lagrange point lattice
function quadrature_composite_lagrange_midpoint(p)
    d_ = 2
    n = binomial(p + 2, 2)
    weights = Vector{Float64}(undef, n)
    points = Matrix{Float64}(undef, 2, n)

    k = 0
    for I in Iterators.product(ntuple(_ -> 0:p, d_)...)
        sum(Tuple(I)) > p && continue
        k += 1
        weights[k] = 1 / n / factorial(d_)
        points[1, k] = I[1] / p
        points[2, k] = I[2] / p
    end

    return weights, points
end
